/*
** Module   : NopPropertyAccessor.java
** Abstract : Do-nothing PropertyAccessor implementation.
**
** Copyright (c) 2004-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description----------------------------
** 001 ECF 20080305 @37507  Created initial version. Do-nothing PropertyAccessor implementation.
** 002 SVL 20120331         Upgraded to Hibernate 4.
** 003 OM  20140411         Replaced hardcoded string computed column signatures with new
**                          DialectHelper case-sensitive aware static constants.
*/

package com.goldencode.p2j.persist;

import java.lang.reflect.*;
import java.util.*;
import com.goldencode.p2j.persist.dialect.*;
import org.hibernate.*;
import org.hibernate.engine.spi.*;
import org.hibernate.property.*;
import com.goldencode.p2j.util.*;

/**
 * Implementation of a do-nothing <code>PropertyAccessor</code>. This is
 * necessary to implement computed column support.  Specifically, computed
 * columns are used in P2J to support embedding the <code>rtrim()</code> and
 * <code>upper()</code> functions in database indexes for text-based index
 * components.
 * <p>
 * If a database dialect requires computed columns, these are injected
 * programatically into the Hibernate ORM documents during server startup.
 * The are set to be read-only, that is, each has its <code>insert</code> and
 * <code>update</code> attribute set to false.  Furthermore, to prevent the
 * need to represent these columns in the DMO classes themselves, this class
 * is used as Hibernate's property accessor, to intercept Hibernate's requests
 * to get and set DMO properties.  This is necessary because the computed
 * column data is only needed by queries at the database level, never in
 * application logic.
 * <p>
 * An instance of this class, when requested to retrieve the accessor's
 * <code>Getter</code>, returns an object which always returns
 * <code>null</code>, and which reports the type of the property which is the
 * basis for the computed column.
 * <p>
 * When the accessor's <code>Setter</code> is requested, an object is returned
 * which simply does nothing when asked to set a property value into a DMO.
 */
public final class NopPropertyAccessor
implements PropertyAccessor
{
   /** Do-nothing <code>Setter</code> instance */
   private static final Setter setter = new NopSetter();

   /**
    * Return a do-nothing <code>Getter</code>.
    * 
    * @param   theClass
    *          DMO type.
    * @param   propertyName
    *          DMO property name.
    * 
    * @return  Do-nothing <code>Getter</code>.
    * 
    * @see org.hibernate.property.PropertyAccessor#getGetter(java.lang.Class, java.lang.String)
    */
   public Getter getGetter(Class theClass, String propertyName)
   throws PropertyNotFoundException
   {
      try
      {
         // TODO: support dialect-specific prefix.
         if (propertyName.startsWith(DialectHelper.SENSITIVE_CHAR_FIELD))
         {
            propertyName = propertyName.substring(DialectHelper.SENSITIVE_CHAR_FIELD.length());
         }
         else if (propertyName.startsWith(DialectHelper.INSENSITIVE_CHAR_FIELD))
         {
            propertyName = propertyName.substring(DialectHelper.INSENSITIVE_CHAR_FIELD.length());
         }
         Class<?> type = theClass.getDeclaredField(propertyName).getType();
         
         return new NopGetter(type);
      }
      catch (NoSuchFieldException exc)
      {
         throw new PropertyNotFoundException(propertyName);
      }
   }

   /**
    * Return a do-nothing <code>Setter</code>.
    * 
    * @param   theClass
    *          DMO type.
    * @param   propertyName
    *          DMO property name.
    * 
    * @return  Do-nothing <code>Setter</code>.
    * 
    * @see org.hibernate.property.PropertyAccessor#getSetter(java.lang.Class, java.lang.String)
    */
   public Setter getSetter(Class theClass, String propertyName)
   throws PropertyNotFoundException
   {
      return setter;
   }
   
   /**
    * Implementation of a do-nothing <code>Getter</code>.  Always returns
    * <code>null</code> when asked for the property value.
    */
   private static class NopGetter
   implements Getter
   {
      /** Property type */
      private final Class<?> type;
      
      /**
       * Constructor which stores property type.
       * 
       * @param   type
       *          DMO property type.
       * 
       * @throws  HibernateException
       *          if <code>type</code> is not assignable from an expected
       *          type (<code>BaseDataType</code>).
       */ 
      NopGetter(Class<?> type)
      {
         this.type = type;
         if (!(BaseDataType.class.isAssignableFrom(type)))
         {
            throw new HibernateException("Unexpected type:  " + type);
         }
      }
      
      /**
       * Get the property value from the given instance.
       * 
       * @param   owner
       *          The instance containing the value to be retrieved.
       * 
       * @return  <code>null</code>, always.
       */
      public Object get(Object owner)
      {
         return null;
      }

      /**
       * Retrieve the member to which this property maps. This might be the field or it might be
       * the getter method.
       *
       * @return  <code>null</code>, always.
       */
      public Member getMember()
      {
         return null;
      }

      /**
       * Get the property value from the given owner instance.
       *
       * @param   owner
       *          The instance containing the value to be retrieved.
       * @param   mergeMap
       *          A map of merged persistent instances to detached instances.
       * @param   session
       *          The session from which this request originated.
       *          
       * @return  <code>null</code>, always.
       */
      public Object getForInsert(Object owner, Map mergeMap, SessionImplementor session)
      throws HibernateException
      {
         return null;
      }

      /**
       * Get the declared Java type.
       * 
       * @return  Declared Java type.
       */
      public Class getReturnType()
      {
         return type;
      }

      /**
       * Optional operation (return null)
       */
      public String getMethodName()
      {
         return null;
      }

      /**
       * Optional operation (return null)
       */
      public Method getMethod()
      {
         return null;
      }
   }
   
   /**
    * Implementation of a do-nothing <code>Setter</code>.
    */
   private static class NopSetter
   implements Setter
   {
      /**
       * Set the property value from the given instance.  This implementation
       * does nothing.
       *
       * @param   target
       *          The instance upon which to set the given value.
       * @param   value
       *          The value to be set on the target.
       * @param   factory
       *          The session factory from which this request originated.
       */
      public void set(Object target, Object value, SessionFactoryImplementor factory)
      {
      }
      
      /**
       * Optional operation (return null)
       */
      public String getMethodName()
      {
         return null;
      }
      
      /**
       * Optional operation (return null)
       */
      public Method getMethod()
      {
         return null;
      }
   }
}
