/*
** Module   : SortCriterion.java
** Abstract : Object which encapsulates a single "order by" sort criterion
**
** Copyright (c) 2004-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description----------------------------
** 001 ECF 20060802   @28346 Created initial version. Based upon an inner
**                           class previously within HQLHelper.
** 002 ECF 20060831   @29089 Added support for sorting by an indexed
**                           property. Specifically, support sorting by
**                           a property of a DMO's associated list of
**                           composite properties.
** 003 ECF 20061120   @31328 Store DMO implementation class. This is the
**                           class represented by a sort property's
**                           qualifying alias in the order by clause.
** 004 ECF 20070216   @32169 Avoid directional mismatch if possible when
**                           appending primary key as a sort criterion.
**                           The primary key sort criterion's direction is
**                           made to match that of the last explicit sort
**                           criterion.
** 005 NVS 20070619   @34102 Fixed IndexOutOfBoundsException in parse()
**                           method where the list was improperly modified
**                           in a loop body.
** 006 ECF 20070705   @34363 Minor optimization. Replaced StringBuffer
**                           with StringBuilder.
** 007 ECF 20070827   @34997 Handle temp table multiplex ID as a sortable
**                           property.
** 008 ECF 20071130   @36125 Fixed redundant 'id' sort criterion. Only
**                           append 'id' as sort criterion if not
**                           redundant with explicit criteria, when
**                           parsing a sort string.
** 009 ECF 20071205   @36262 Added toRawSortPhrase() method. This is used
**                           by SortIndex when initializing a new
**                           instance of that class.
** 010 ECF 20071218   @36462 Fixed handling of indexed sort criteria.
**                           toRawSortPhrase() was producing an invalid
**                           phrase in cases of indexed sort criteria.
** 011 ECF 20080310   @37486 Added support for computed columns. Computed
**                           column property names are injected to replace
**                           normal character property names as necessary.
** 012 ECF 20080330   @37739 Modified to support method change in
**                           DatabaseManager.
** 013 SVL 20080418   @38035 makeUnique parameter added to parse().
**                           doSpecificProcessing parameter added to
**                           toSortExpression().
** 014 ECF 20080509   @38254 Fixed regression introduced by #011 (@37486).
**                           Composite qualifiers were not being handled
**                           correctly when generating sort and where
**                           expressions.
** 015 ECF 20080510   @38502 Refactored to enable access from sub-package.
**                           Made class and some methods public. Added new
**                           parse() and ctor variants.
** 016 ECF 20080708   @39076 Added isPrimaryKey() method. Indicates if
**                           sort criterion represents the backing table's
**                           primary key.
** 017 ECF 20080924   @39937 Memory tuning. intern()'d strings which are
**                           re-used within and across sessions.
** 018 SVL 20140210          Use HQLExpression instead of HQL string.
** 019 OM  20140410          Fixed support case-sensitive fields in indexes.
*/

package com.goldencode.p2j.persist;

import java.lang.reflect.*;
import java.util.*;
import com.goldencode.p2j.persist.dialect.*;
import com.goldencode.p2j.util.*;

/**
 * An individual sort criterion, consisting of a property name, sort
 * direction, DMO implementation class associated with an alias qualifier,
 * getter method for the property, and a flag indicating whether case is
 * ignored when sorting a <code>character</code> property.  Provides static
 * convenience methods to parse sort criteria and to produce an HQL-compliant
 * order by clause.
 * <p>
 * Two forms of the property name are stored:  the <code>name</code> and the
 * <code>orginalName</code>.  For non-indexed properties, these will be the
 * same.  For indexed properties, <code>name</code> is the expanded name
 * after considering the ANSI join needed to support the sort criterion;
 * <code>originalName</code> is the name as originally given before parsing.
 * For instance, the sort criterion <code>person.schedule[0] asc</code> would
 * have the following <code>name</code>:
 * <pre>
 *    person_composite5.schedule
 * </pre>
 * and the following <code>originalName<code>:
 * <pre>
 *    person.schedule[0]
 * </pre>
 * <p>
 * The from clause would be augmented with
 * <pre>
 *    ...join person.composite5 as person_composite5...
 * </pre>
 * and the where clause in this case would be augmented with
 * <pre>
 *    ...[and/where] index(person_composite5) = 0...
 * </pre>
 * <p>
 * Thus, given a DMO with a <code>schedule</code> field of extent 5, a basic
 * HQL statement of
 * <pre>
 *    from Person as person order by person.schedule[0] asc
 * </pre>
 * would be expanded to
 * <pre>
 *    from PersonImpl as person
 *    join person.composite5 as person_composite5
 *    where (index(person_composite5) = 0)
 *    order by person.empNum asc,
 *             upper(rtrim(person_composite5.schedule, ' \t\n\r')) asc
 * <pre>
 * (assuming <code>schedule</code> is a case-insensitive character field of
 * extent 5).
 * <p>
 * Note that this class does not actually refactor the full HQL statement;
 * it provides the sort criteria helper methods that allow the various query
 * implementations to do so.
 */
public final class SortCriterion
{
   /** <code>true</code> if sort is ascending, else it is descending */
   private final boolean ascending;
   
   /** DMO implementation class containing the sort property */
   private final Class<?> dmoClass;
   
   /** Getter method associated with sort property */
   private final Method method;
   
   /** Is property of type character? */
   private final boolean isCharacter;
   
   /** Is case ignored (for character property only)? */
   private final boolean ignoreCase;
   
   /** Computed column prefix for property name (optional) */
   private final String computedColumnPrefix;
   
   /** DMO alias qualifier, if any */
   private final String alias;
   
   /** Unqualified property name */
   private final String propertyName;
   
   /** Sort property name, including subscript if any, usually qualified */
   private String originalName;
   
   /** Sort property name, usually qualified */
   private String name;
   
   /** Name of composite alias, if any */
   private String composite = null;
   
   /** Subscript value, if any */
   private int subscript = -1;
   
   /**
    * Constructor which parses original sort component text, deriving the
    * property name, sort direction, getter method for the property, case
    * sensitivity of the property (character type only), and possibly a
    * subscript value (for extent fields) and a composite alias name (for
    * extent fields which are converted to use composite inner classes within
    * the DMO).
    * 
    * @param   buffer
    *          Record buffer.
    * @param   text
    *          Text representation of sort component, as provided by query.
    * @param   schema
    *          Name of database schema.
    * @param   dmoIface
    *          DMO interface.
    * @param   dmoClass
    *          DMO implementation class.
    *
    * @throws  PersistenceException
    *          if there is any error with the text representation of the sort
    *          component provided by the caller.
    */
   SortCriterion(RecordBuffer buffer,
                 String text,
                 String schema,
                 Class<?> dmoIface,
                 Class<?> dmoClass)
   throws PersistenceException
   {
      this(buffer.getDialect(), text, schema, dmoIface, dmoClass);
   }
   
   /**
    * Constructor which parses original sort component text, deriving the
    * property name, sort direction, getter method for the property, case
    * sensitivity of the property (character type only), and possibly a
    * subscript value (for extent fields) and a composite alias name (for
    * extent fields which are converted to use composite inner classes within
    * the DMO).
    * 
    * @param   dialect
    *          Database dialect.
    * @param   text
    *          Text representation of sort component, as provided by query.
    * @param   schema
    *          Name of database schema.
    * @param   dmoIface
    *          DMO interface.
    * @param   dmoClass
    *          DMO implementation class.
    *
    * @throws  PersistenceException
    *          if there is any error with the text representation of the sort
    *          component provided by the caller.
    */
   SortCriterion(P2JDialect dialect,
                 String text,
                 String schema,
                 Class<?> dmoIface,
                 Class<?> dmoClass)
   throws PersistenceException
   {
      this.dmoClass = dmoClass;
      String[] tmp = text.split(" ");
      String rawName = tmp[0].trim();
      name = rawName;
      originalName = rawName.intern();
      int startPos = rawName.indexOf("[");
      int endPos = -1;
      if (startPos >= 0)
      {
         endPos = rawName.indexOf("]");
         if (endPos < 0)
         {
            throw new PersistenceException(
               "Malformed subscript text:  " + text);
         }
         
         // Extract base name.
         name = rawName.substring(0, startPos);
      }
      
      int dot = name.lastIndexOf(".");
      if (dot < 0)
      {
         throw new PersistenceException(
            "Sort property may not be unqualified:  " + rawName);
      }
      this.propertyName = name.substring(dot + 1).intern();
      this.alias = name.substring(0, dot).intern();
      
      if (startPos >= 0)
      {
         // Check Hibernate configuration to see if we match a composite for
         // this alias and property combination.  Will resolve to null if
         // there is none.
         composite = DatabaseManager.getComposite(dmoClass, propertyName);
         if (composite != null)
         {
            // Extract subscript.
            String sub = rawName.substring(startPos + 1, endPos);
            try
            {
               subscript = Integer.parseInt(sub);
            }
            catch (NumberFormatException exc)
            {
               throw new PersistenceException(
                  "Unsupported subscript:  " + sub);
            }
            
            // Compose full composite name, including DMO alias.
            StringBuilder buf = new StringBuilder(alias);
            buf.append("_");
            buf.append(composite);
            composite = buf.toString().intern();
            buf.append(".");
            buf.append(propertyName);
            name = buf.toString();
         }
      }
      
      name = name.intern();
      
      // Determine sort direction.
      ascending = (tmp.length == 1 || !"desc".equalsIgnoreCase(tmp[1].trim()));
      
      // Get backing getter method.
      if (TemporaryBuffer.MULTIPLEX_FIELD_NAME.equals(propertyName))
      {
         method = null;
         isCharacter = false;
         ignoreCase = false;
         computedColumnPrefix = null;
      }
      else
      {
         String ccPrefix = null;
         Class<?>[] parms = (startPos >= 0
                             ? new Class<?>[] { Integer.TYPE }
                             : new Class<?>[] {});
         method = DBUtils.findDMOGetterMethod(dmoClass, propertyName, parms);
         
         // Determine case sensitivity and computed column prefix, if needed.
         boolean caseSens = false;
         isCharacter = character.class.equals(method.getReturnType());
         if (isCharacter)
         {
            Boolean ccIgnoreCase = null;
            if (dialect.needsComputedColumns())
            {
               ccIgnoreCase = DatabaseManager.getIgnoreCase(dmoClass, propertyName);
            }
            if (ccIgnoreCase != null)
            {
               caseSens = !ccIgnoreCase;
               ccPrefix = dialect.getComputedColumnPrefix(caseSens);
            }
            else
            {
               caseSens = DMOIndex.isCaseSensitive(schema, dmoIface, propertyName);
            }
            if (ccPrefix != null)
            {
               ccPrefix = ccPrefix.intern();
            }
         }
         ignoreCase = isCharacter && !caseSens;
         computedColumnPrefix = ccPrefix;
      }
   }
   
   /**
    * Parse a sort phrase (not including the &quot;order by&quot; preamble)
    * into a list of <code>SortCriterion</code> objects.  All components of
    * the phrase are assumed to be properties of the same DMO, which is the
    * DMO associated with <code>buffer</code>.
    *
    * @param   dialect
    *          Database dialect.
    * @param   sort
    *          Sort phrase.
    * @param   schema
    *          Database schema.
    * @param   dmoAlias
    *          Alias qualifier which represents DMO in sort phrase.
    * @param   dmoClass
    *          DMO implementation class.
    * @param   dmoIface
    *          DMO interface.
    * @param   makeUnique
    *          Indicates whether the <code>&lt;dmoAlias&gt;.id</code>
    *          criterion should be appended to the returned criteria list. 
    *
    * @return  List of <code>SortCriterion</code> objects.
    *
    * @throws  PersistenceException
    *          if there is any error parsing the given string.
    */
   public static List<SortCriterion> parse(P2JDialect dialect,
                                           String sort,
                                           String schema,
                                           String dmoAlias,
                                           Class<?> dmoClass,
                                           Class<?> dmoIface,
                                           boolean makeUnique)
   throws PersistenceException
   {
      boolean pKeyAscend = true;
      
      // Parse sort clause (if any).
      List<SortCriterion> sortCriteria = new LinkedList<SortCriterion>();
      Set<String> sortNames = new HashSet<String>();
      if (sort != null && sort.trim().length() > 0)
      {
         SortCriterion next = null;
         String[] tmp = sort.split(",");
         for (int i = 0; i < tmp.length; i++)
         {
            next = new SortCriterion(dialect,
                                     tmp[i].trim(),
                                     schema,
                                     dmoIface,
                                     dmoClass);
            sortCriteria.add(next);
            sortNames.add(next.getUnqualifiedName());
         }
         
         // Match the primary key sort direction to that of the last sort
         // criterion.
         if (next != null)
         {
            pKeyAscend = next.isAscending();
         }
      }
      
      // Tentatively add primary key as the last component of the sort
      // criteria, if not redundant with the previous criteria.
      if (makeUnique && !sortNames.contains("id"))
      {
         StringBuilder buf = new StringBuilder();
         buf.append(dmoAlias);
         buf.append(".id ");
         buf.append(pKeyAscend ? "asc" : "desc");
         sortCriteria.add(new SortCriterion(dialect,
                                            buf.toString(),
                                            schema,
                                            dmoIface,
                                            dmoClass));
      }
      
      // Eliminate superfluous sort components if sort criteria define a
      // superset of a unique constraint.
      int size = sortCriteria.size();
      int lastUniqueComp = size;
      Iterator<Set<String>> uniques =
         DMOIndex.uniqueConstraints(schema, dmoIface);
      while (uniques.hasNext())
      {
         Set<String> compSet = uniques.next();
         if (sortNames.containsAll(compSet))
         {
            ListIterator<SortCriterion> iter =
               sortCriteria.listIterator(size);
            for (int j = size - 1; iter.hasPrevious(); j--)
            {
               SortCriterion prev = iter.previous();
               String name = prev.getUnqualifiedName();
               if (!compSet.contains(name))
               {
                  lastUniqueComp = Math.min(lastUniqueComp, j);
                  continue;
               }
               
               break;
            }
         }
         sortCriteria = sortCriteria.subList(0, lastUniqueComp);
         size = sortCriteria.size();
      }
      
      return sortCriteria;
   }
   
   /**
    * Parse a sort phrase (not including the &quot;order by&quot; preamble)
    * into a list of <code>SortCriterion</code> objects.  All components of
    * the phrase are assumed to be properties of the same DMO, which is the
    * DMO associated with <code>buffer</code>.
    *
    * @param   sort
    *          Sort phrase.
    * @param   buffer
    *          Record buffer associated with the sort phrase.
    *
    * @return  List of <code>SortCriterion</code> objects.  This list should
    *          never be empty;  even if <code>sort</code> is null or empty,
    *          a single criterion of <code>&lt;dmoAlias&gt;.id</code> will be
    *          created.
    * 
    * @throws  PersistenceException
    *          if there is any error parsing the given string.
    */
   public static List<SortCriterion> parse(String sort, RecordBuffer buffer)
   throws PersistenceException
   {
      return parse(sort, buffer, true);
   }

   /**
    * Parse a sort phrase (not including the &quot;order by&quot; preamble)
    * into a list of <code>SortCriterion</code> objects.  All components of
    * the phrase are assumed to be properties of the same DMO, which is the
    * DMO associated with <code>buffer</code>.
    *
    * @param   sort
    *          Sort phrase.
    * @param   buffer
    *          Record buffer associated with the sort phrase.
    * @param   makeUnique
    *          Indicates whether the <code>&lt;dmoAlias&gt;.id</code>
    *          criterion should be appended to the returned criteria list. 
    *
    * @return  List of <code>SortCriterion</code> objects.
    *
    * @throws  PersistenceException
    *          if there is any error parsing the given string.
    */
   static List<SortCriterion> parse(String sort,
                                    RecordBuffer buffer,
                                    boolean makeUnique)
   throws PersistenceException
   {
      P2JDialect dialect = buffer.getDialect();
      String schema = buffer.getSchema();
      String dmoAlias = buffer.getDMOAlias();
      Class<?> dmoClass = buffer.getDMOImplementationClass();
      Class<?> dmoIface = buffer.getDMOInterface();
      
      return parse(dialect,
                   sort,
                   schema,
                   dmoAlias,
                   dmoClass,
                   dmoIface,
                   makeUnique);
   }
   
   /**
    * Assemble an HQL &quot;order by&quot; clause, sans the &quot;order
    * by&quot; preamble, from the given list of sort criteria, optionally
    * inverting the sort direction of the sort criteria.
    *
    * @param   sortCriteria
    *          List of <code>SortCriterion</code> objects.
    * @param   invert
    *          <code>true</code> to reverse the sort direction of all sort
    *          criteria;  <code>false</code> to use the original sort
    *          directions of the sort criteria.
    *
    * @return  HQL order by clause, sans preamble.
    */
   static String assembleHQLNoPreamble(List<SortCriterion> sortCriteria,
                                       boolean invert)
   {
      StringBuilder buf = new StringBuilder();
      Iterator<SortCriterion> iter = sortCriteria.iterator();
      for (int i = 0; iter.hasNext(); i++)
      {
         SortCriterion next = iter.next();
         if (i > 0)
         {
            buf.append(", ");
         }
         buf.append(next.toSortExpression(invert, false));
      }
      
      return buf.toString();
   }
   
   /**
    * Generate a raw sort phrase from the given list of criteria, in the
    * simple form:
    * <pre>
    *    {name} {direction} [, {name} {direction}]
    * </pre>
    * <p>
    * Character fields are not enclosed in any upper() or rtrim() functions.
    * 
    * @param   sortCriteria
    *          List of sort criteria.
    *          
    * @return  Raw sort phrase.
    */
   static String toRawSortPhrase(List<SortCriterion> sortCriteria)
   {
      StringBuilder buf = new StringBuilder();
      Iterator<SortCriterion> iter = sortCriteria.iterator();
      for (int i = 0; iter.hasNext(); i++)
      {
         SortCriterion next = iter.next();
         if (i > 0)
         {
            buf.append(", ");
         }
         buf.append(next.getOriginalName());
         buf.append(next.isAscending() ? " asc" : " desc");
      }
      
      return buf.toString();
   }
   
   /**
    * Get the unqualified name of the property backing this component.
    *
    * @return  Property name, sans DMO alias qualifier.
    */
   public String getUnqualifiedName()
   {
      return propertyName;
   }
   
   /**
    * Get a normalized string representation of the sort component.  Uses
    * the form:
    * <pre>
    *   {name} {direction}
    * </pre>
    * <p>
    * For example:
    * <pre>
    *   customer.accountNumber asc
    * </pre>
    * <p>
    * If this component represents a case-insensitive sort on a character
    * property, the property name will be enclosed in the
    * <code>upper</code> database function, as in:
    * <pre>
    *   upper(customer.name) asc
    * </pre>
    *
    * @return  String representation as described above.
    */
   public String toString()
   {
      return toSortExpression(false);
   }
   
   /**
    * Indicate whether this sort criterion represents the DMO's primary key.
    * 
    * @return  <code>true</code> if this sort criterion represents the DMO's
    *          primary key, else <code>false</code>.
    */
   boolean isPrimaryKey()
   {
      return DatabaseManager.PRIMARY_KEY.equals(propertyName);
   }
   
   /**
    * Get the name of the property behind this component.
    * <p>
    * This will usually be the same as the name returned by {@link
    * #getName()}, except in the case of an extent field.  In that case, this
    * method will return the name with the original subscripting.
    *
    * @return  Property name, usually qualified by DMO alias.
    */
   String getOriginalName()
   {
      return originalName;
   }
   
   /**
    * Get the name of the property behind this component.
    * <p>
    * This will usually be the same as the name returned by {@link
    * #getOriginalName()}, except in the case of an extent field.  In that
    * case, this method will return the expanded name, in the form:
    * <pre>
    *    {alias}_composite{num}_{original_name_sans_subscript}
    * </pre>
    *
    * @return  Property name, usually qualified by DMO alias.
    */
   String getName()
   {
      return name;
   }
   
   /**
    * Indicate whether component sorts in ascending order.
    *
    * @return  <code>true</code> if ascending, <code>false</code> if
    *          descending.
    */
   boolean isAscending()
   {
      return ascending;
   }
   
   /**
    * Get the DMO implementation class containing the sort property for this
    * criterion.
    *
    * @return  DMO implementation class.
    */
   Class<?> getDMOClass()
   {
      return dmoClass;
   }
   
   /**
    * Get the getter method associated with the property behind this
    * component.
    *
    * @return  Getter method.
    */
   Method getMethod()
   {
      return method;
   }
   
   /**
    * Get a normalized string representation of the sort component.  Uses
    * the form:
    * <pre>
    *   {name} {direction}
    * </pre>
    * <p>
    * For example:
    * <pre>
    *   customer.accountNumber asc
    * </pre>
    * <p>
    * If this component represents a case-insensitive sort on a character
    * property, the property name will be enclosed in the
    * <code>upper</code> database function, as in:
    * <pre>
    *   upper(customer.name) asc
    * </pre>
    * <p>
    * If the <code>invert</code> parameter is <code>true</code>, the sort
    * direction will be reversed, as in:
    * <pre>
    *   upper(customer.name) desc
    * </pre>
    *
    * @param   invert
    *          <code>true</code> to reverse the sort direction.
    *
    * @return  String representation as described above.
    */
   String toSortExpression(boolean invert)
   {
      return toSortExpression(invert, true);
   }
   
   /**
    * Get a normalized string representation of the sort component.  Uses
    * the form:
    * <pre>
    *   {name} {direction}
    * </pre>
    * <p>
    * For example:
    * <pre>
    *   customer.accountNumber asc
    * </pre>
    * <p>
    * If this component represents a case-insensitive sort on a character
    * property and <code>doSpecificProcessing</code> is set, the property name
    * will be enclosed in the <code>upper</code> database function, as in:
    * <pre>
    *   upper(customer.name) asc
    * </pre>
    * <p>
    * If the <code>invert</code> parameter is <code>true</code>, the sort
    * direction will be reversed, as in:
    * <pre>
    *   upper(customer.name) desc
    * </pre>
    *
    * @param   invert
    *          <code>true</code> to reverse the sort direction.
    *
    * @param   doSpecificProcessing
    *          <code>true</code> to perform dialect-specific processing
    *          (adding computed column prefix, string trimming and
    *          uppercasing)
    *
    * @return  String representation as described above.
    */
   String toSortExpression(boolean invert, boolean doSpecificProcessing)
   {
      StringBuilder buf = new StringBuilder();
      
      if (!doSpecificProcessing)
      {
         buf.append(composite != null ? name : originalName);
      }
      else
      {
         DBUtils.composePropertyName(isCharacter,
                                     ignoreCase,
                                     computedColumnPrefix,
                                     (composite != null ? composite : alias),
                                     propertyName,
                                     buf);
      }

      buf.append(" ");
      
      if (invert)
      {
         buf.append(ascending ? "desc" : "asc");
      }
      else
      {
         buf.append(ascending ? "asc" : "desc");
      }
      
      return buf.toString();
   }
   
   /**
    * If it is necessary to perform a join to support this sort criterion,
    * return the necessary join clause, in ANSI join format.  This is needed
    * to support a sort by a specific element of an array property, where the
    * property resides within a composite inner class of a DMO.
    * <p>
    * This will be in the form:
    * <pre>
    *    join {alias}.{composite} as {alias}_{composite}
    * </pre>
    * for instance:
    * <pre>
    *    join workWeek.composite5 as workWeek_composite5
    * </pre>
    *
    * @return  ANSI join clause if needed, or <code>null</code> if not needed.
    */
   String getAnsiJoin()
   {
      if (composite == null)
      {
         return null;
      }
      
      StringBuilder buf = new StringBuilder("join ");
      buf.append(composite.replace('_', '.'));
      buf.append(" as ");
      buf.append(composite);
      
      return buf.toString();
   }
   
   /**
    * If it is necessary to perform a join to support this sort criterion,
    * return a where clause component which will restrict the joined
    * association (a list) to the intended list element.  This is needed to
    * support a sort by a specific element of an array property, where the
    * property resides within a composite inner class of a DMO.
    * <p>
    * This will be in the form:
    * <pre>
    *    index({alias}_{composite}) = {subscript}
    * </pre>
    * for instance:
    * <pre>
    *    index(workWeek_composite5) = 2
    * </pre>
    *
    * @return  Where clause restrictor component, or <code>null</code> if not
    *          needed.
    */
   String getCompositeRestrictor()
   {
      if (composite == null)
      {
         return null;
      }
      
      StringBuilder buf = new StringBuilder("index(");
      buf.append(composite);
      buf.append(") = ");
      buf.append(subscript);
      
      return buf.toString();
   }
   
   /**
    * Get a string representation of this sort component which is suitable
    * to append to a base where clause when augmenting that where clause
    * to make it more specific.  The string represents an equality or
    * range match between the property name and a substitution parameter.
    * A range match may be inverted to represent the opposite sort
    * direction.  For example, the sort component
    * <pre>
    *   customer.accountNumber asc
    * </pre>
    * would be represented as:
    * <pre>
    *   customer.accountNumber = ?
    * </pre>
    * for an equality match;  as:
    * <pre>
    *   customer.accountNumber > ?
    * </pre>
    * for a normal range match;  and as:
    * <pre>
    *   customer.accountNumber < ?
    * </pre>
    * for an inverted range match.
    * <p>
    * Case-insensitive character properties are wrapped in the
    * <code>upper</code> database function, as in:
    * <pre>
    *   upper(customer.name) = ?
    * </pre>
    *
    * @param   rangeMatch
    *          <code>true</code> if a range match expression should be
    *          created;  <code>false</code> to create an equality match
    *          expression.
    * @param   invert
    *          <code>true</code> to reverse polarity of a range match
    *          expression, else <code>false</code>.
    *
    * @return  Where clause match expression as described above.
    */
   HQLExpression toWhereExpression(boolean rangeMatch, boolean invert)
   {
      StringBuilder buf = new StringBuilder();
      DBUtils.composePropertyName(isCharacter,
                                  ignoreCase,
                                  computedColumnPrefix,
                                  (composite != null ? composite : alias),
                                  propertyName,
                                  buf);

      HQLExpression whereClause = new HQLExpression(buf.toString());

      if (rangeMatch)
      {
         if (invert)
         {
            whereClause.append(ascending ? " < " : " > ");
         }
         else
         {
            whereClause.append(ascending ? " > " : " < ");
         }
      }
      else
      {
         whereClause.append(" = ");
      }

      if (ignoreCase)
      {
         whereClause.append("upper(", true);
         whereClause.append(")");
      }
      else
      {
         whereClause.append("", true);
      }

      return whereClause;
   }
}
